'use strict';

require('./curry.js');
var isString = require('./isString-d8679d99.js');
require('./flatten-a717674f.js');
require('./variadic.js');
require('./_append.js');
require('./append.js');
require('./flip.js');
require('./concatWith-aa95abaf.js');
require('./isArray-fff5bad6.js');
var mempty = require('./mempty-fb137ec7.js');
var concat = require('./concat.js');
var empty = require('./empty.js');
var head = require('./head.js');
var tail = require('./tail.js');
require('./_prepend.js');
var _foldRight = require('./_foldRight.js');
var prepend = require('./prepend.js');

/**
 * Returns `a` in a pure context.
 *
 * @private
 */

function pure(a) {
  return isString.isString(a) || mempty.isArrayOfStrings(a) ? a : [a];
}

/**
 * Calculates the subsequences of a list.
 *
 * @param {Array|String} as The list.
 * @returns {Array} A list that contains all the subsequences of the elements
 * in the list of `as`.
 * @example
 *
 * import { subsequences } from 'fkit'
 * subsequences([1, 2, 3]) // [[], [1], [2], [1, 2], [3], [1, 3], [2, 3], [1, 2, 3]]
 * subsequences('abc') // ['', 'a', 'b', 'ab', 'c', 'ac', 'bc', 'abc']
 */

function subsequences(as) {
  var subsequences_ = function subsequences_(bs) {
    var b = head(bs);

    var f = function f(ys, r) {
      return concat(pure(ys), pure(prepend(b, ys)), r);
    };

    if (empty(bs)) {
      return [];
    } else {
      return prepend(pure(b), _foldRight(f, [], subsequences_(tail(bs))));
    }
  };

  return prepend(mempty.mempty(as), subsequences_(as));
}

module.exports = subsequences;
